% Kernel PCA toy example for k(x,y)=exp(-||x-y||^2/rbf_var), cf. Fig. 4 in 
% @article{SmoManSch99,
%   author    = "A.~Smola and O.~Mangasarian and B.~{Sch\"olkopf}",
%   title     = "Sparse Kernel Feature Analysis",
%   journal =	 {Neural Computation, submitted},
%   volume    = ,
%   issue     = ,
%   pages     = "",
%   year      = 1999}
% This file can be downloaded from http://svm.first.gmd.de.
% Last modified: 15 August 1999 (smola)


% parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rbf_var = 0.1;
xnum = 4;
ynum = 1;
max_feat = xnum*ynum;
% (extract features from the first <max_feat> Eigenvectors)
x_test_num = 15;
y_test_num = 15;
cluster_pos = [-0.5 -0.2; 0 0.6; 0.5 0];
cluster_size = 30;

maxtail_on = 0;				% turn on maxtail version
maxtail_subset = 10;			% try out 10 directions 
                                        % (i.e. 10/11 quantile on average)

renormalize_on = 0;			% renormalize alphas

rand('seed',0);				% reproducible results

% generate a toy data set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
num_clusters = size(cluster_pos,1);
train_num = num_clusters*cluster_size;
patterns = zeros(train_num, 2);
range = 1.2;
randn('seed', 0);
for i=1:num_clusters,
  patterns((i-1)*cluster_size+1:i*cluster_size,1) = cluster_pos(i,1)+0.1*randn(cluster_size,1);
  patterns((i-1)*cluster_size+1:i*cluster_size,2) = cluster_pos(i,2)+0.1*randn(cluster_size,1);
end
test_num = x_test_num*y_test_num;
x_range = -range:(2*range/(x_test_num - 1)):range;
%y_offset = 0.5;
y_offset = 0;
y_range = -range+ y_offset:(2*range/(y_test_num - 1)):range+ y_offset;
[xs, ys] = meshgrid(x_range, y_range);
test_patterns(:, 1) = xs(:);
test_patterns(:, 2) = ys(:);
cov_size = train_num;  % use all patterns to compute the covariance matrix

% carry out Sparse KFA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for i=1:cov_size,
  for j=i:cov_size,
    K(i,j) = exp(-norm(patterns(i,:)-patterns(j,:))^2/rbf_var);
    K(j,i) = K(i,j);
  end
end
unit = ones(cov_size, cov_size)/cov_size;

alpha_zero = ones(cov_size,1);	        % equal weight for all features
alpha = zeros(cov_size,max_feat);	% temporary directions
alpha_feat = zeros(max_feat);		% at most a triangle of size cov_size
idx = [];				% this will contain all the
                                        % kernel functions 
					
Q = zeros(cov_size,1);			% the q values (variance
                                        % etc.), initialize it for
					% speed
Q_feat = zeros(max_feat,1);		% best features

for i=1:max_feat
  if (i > 1)
    projections = K .* (alpha_zero * ones(1,cov_size)) + ...
	alpha(:,1:(i-1)) * K(idx,:);
  else
    projections = K .* (alpha_zero * ones(1,cov_size));
  end    
  % compute variance, etc
  for j=1:cov_size
    Q(j) = std(projections(j,:));	% does centering automatically
  end;
  Q(idx) = zeros(1,i-1);			% blank out used directions

  if (maxtail_on == 0)
    [max_q, max_idx] = max(Q);
  else
    % randomized selection (the n/n+1 quantile trick), just a dirty hack 
    % for proof of concept  
    % remove the patterns in the index (dumb way to do it, i know)
    temp = 1:cov_size;
    temp(idx) = zeros(1,i-1);
    temp = -sort(-temp);
    temp = temp(1:(cov_size - i + 1));
    [dummy, tempindex] = sort(rand(1,(cov_size - i + 1)));
    temp = temp(tempindex(1:maxtail_subset));
  
    [max_q, max_idx] = max(Q(temp));
    max_idx = temp(max_idx);
  end;

  % store the entries
  if (i > 1)
    alpha_feat(i,1:(i-1)) = alpha(max_idx,1:(i-1));
  end;
  alpha_feat(i,i) = alpha_zero(max_idx);
  idx(i) = max_idx;			% that's the new index we're
                                        % looking for
  Q_feat(i) = max_q;					
  			
  % renormalize the directions (and project out old one)
  K_sub = K(idx, idx);			% a submatrix we'll need quite often
  alpha_sub = alpha_feat(i,1:i);		% the new feature vector;
  phisquare = alpha_sub * K_sub * alpha_sub';
  dotprod = alpha_zero .* (K(:,idx) * alpha_sub') + ...
      alpha(:,1:i) * (K_sub * alpha_sub');
  dotprod = dotprod / phisquare;	% the update factors
%  alpha(:,1:i) = alpha(:,1:i) - dotprod * alpha_sub;

  
  if (renormalize_on == 1)
    % if we want to keep l_1 normalization we have to do the following
    sumalpha = alpha_zero + sum(abs(alpha)')';
    alpha_zero = alpha_zero ./ sumalpha;
    alpha = alpha ./ (sumalpha * ones(1,max_feat));
  end;
end;  

% extract features
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  do not need the following here - only need test point features
%unit_train = ones(train_num,cov_size)/cov_size;
%for i=1:train_num,
%  for j=1:cov_size,
%    K_train(i,j) = exp(-norm(patterns(i,:)-patterns(j,:))^2/rbf_var);
%  end
%end
%K_train_n = K_train - unit_train*K - K_train*unit + unit_train*K*unit;
%features = zeros(train_num, max_feat);
%features = K_train_n * evecs(:,1:max_feat);

unit_test = ones(test_num,cov_size)/cov_size;
K_test = zeros(test_num,max_feat);
for i=1:test_num,
  for j=1:max_feat,
    K_test(i,j) = exp(-norm(test_patterns(i,:)-patterns(idx(j),:))^2/rbf_var);
  end
end
%K_test_n = K_test - unit_test*K - K_test*unit + unit_test*K*unit;
test_features = K_test * alpha_feat';
test_features = test_features - ones(test_num,1) * mean(test_features);

% plot it
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
figure(1); clf
for n = 1:max_feat,
  subplot(ynum, xnum, n);
  axis([-range range -range+y_offset range+y_offset]);
  imag = reshape(test_features(:,n), y_test_num, x_test_num);
  axis('xy')
  colormap(gray);
  hold on;
  pcolor(x_range, y_range, imag);
  shading interp
  contour(x_range, y_range, imag, 9, 'b');
  plot(patterns(:,1), patterns(:,2), 'r.')
  text(-1,1.65,sprintf('Q-Value =%4.3f', Q_feat(n)));
  hold off;
end






